# This file is part of colorspacious
# Copyright (C) 2014 Nathaniel Smith <njs@pobox.com>
# See file LICENSE.txt for license information.

import numpy as np

__all__ = ["standard_illuminant_XYZ100", "as_XYZ100_w"]

# The "standard" 2 degree observer (CIE 1931)
# Sourced from http://www.easyrgb.com/index.php?X=MATH&H=15
cie1931 = {
    "A":   [109.850, 100,  35.585],
    "C":   [ 98.074, 100, 118.232],
    "D50": [ 96.422, 100,  82.521],
    "D55": [ 95.682, 100,  92.149],
    # This is sRGB's standard whitepoint:
    "D65": [ 95.047, 100, 108.883],
    "D75": [ 94.972, 100, 122.638],
}

# The "supplementary" 10 degree observer (CIE 1964)
# Sourced from http://www.easyrgb.com/index.php?X=MATH&H=15
cie1964 = {
    "A":   [111.144, 100,  35.200],
    "C":   [ 97.285, 100, 116.145],
    "D50": [ 96.720, 100,  81.427],
    "D55": [ 95.799, 100,  90.926],
    "D65": [ 94.811, 100, 107.304],
    "D75": [ 94.416, 100, 120.641],
}

# Fairchild (2005) says: "The difference between the two standard observers is
# significant, so care should be taken to report which observer is used with
# any colorimetric data. The differences are computationally significant, but
# certainly within the variability of color matching functions found for
# either 2 degree or 10 degree visual fields. Thus the two standard
# colorimetric observers can be thought of as representing the color matching
# functions of two individuals." (page 77)
#
# Though he also says: "It should also be noted that for sample sizes greater
# than 4 degrees, use of the CIE 1964 supplementary standard colorimetric
# observer is recommended." (page 82, when discussing CIE94 delta-E)

def standard_illuminant_XYZ100(name, observer="CIE 1931 2 deg"):
    """Takes a string naming a standard illuminant, and returns its XYZ
    coordinates (normalized to Y = 100).

    We currently have the following standard illuminants in our database:

    * ``"A"``
    * ``"C"``
    * ``"D50"``
    * ``"D55"``
    * ``"D65"``
    * ``"D75"``

    If you need another that isn't on this list, then feel free to send a pull
    request.

    When in doubt, use D65: it's the whitepoint used by the sRGB standard
    (61966-2-1:1999) and ISO 10526:1999 says "D65 should be used in all
    colorimetric calculations requiring representative daylight, unless there
    are specific reasons for using a different illuminant".

    By default, we return points in the XYZ space defined by the CIE 1931 2
    degree standard observer. By specifying ``observer="CIE 1964 10 deg"``,
    you can instead get the whitepoint coordinates in XYZ space defined by the
    CIE 1964 10 degree observer. This is probably only useful if you have XYZ
    points you want to do calculations on that were somehow measured using the
    CIE 1964 color matching functions, perhaps via a spectrophotometer;
    consumer equipment (monitors, cameras, etc.) assumes the use of the CIE
    1931 standard observer in all cases I know of.

    """
    if observer == "CIE 1931 2 deg":
        return np.asarray(cie1931[name], dtype=float)
    elif observer == "CIE 1964 10 deg":
        return np.asarray(cie1964[name], dtype=float)
    else:
        raise ValueError("observer must be 'CIE 1931 2 deg' or "
                         "'CIE 1964 10 deg', not %s" % (observer,))

def test_standard_illuminant_XYZ100():
    assert np.allclose(
        standard_illuminant_XYZ100("D65"),
        [ 95.047, 100, 108.883])

    assert np.allclose(
        standard_illuminant_XYZ100("D65", observer="CIE 1964 10 deg"),
        [ 94.811, 100, 107.304])

    from nose.tools import assert_raises
    assert_raises(ValueError, standard_illuminant_XYZ100, "D65",
                  observer="something else")

# Convenience function
def as_XYZ100_w(whitepoint):
    """A convenience function for getting whitepoints.

    ``whitepoint`` can be either a string naming a standard illuminant (see
    :func:`standard_illuminant_XYZ100`), or else a whitepoint given explicitly
    as an array-like of XYZ values.

    We internally call this function anywhere you have to specify a whitepoint
    (e.g. for CIECAM02 or CIELAB conversions).

    Always uses the "standard" 2 degree observer.

    """
    if isinstance(whitepoint, str):
        return standard_illuminant_XYZ100(whitepoint)
    else:
        whitepoint = np.asarray(whitepoint, dtype=float)
        if whitepoint.shape[-1] != 3:
            raise ValueError("Bad whitepoint shape")
        return whitepoint

def test_as_XYZ100_w():
    assert np.allclose(as_XYZ100_w("D65"), [ 95.047, 100, 108.883])
    assert np.allclose(as_XYZ100_w([1, 2, 3]), [1, 2, 3])
    assert as_XYZ100_w([1, 2, 3]).dtype == float

    from nose.tools import assert_raises
    assert_raises(KeyError, as_XYZ100_w, "D666")
    assert_raises(ValueError, as_XYZ100_w, [1, 2, 3, 4])



# This file is part of colorspacious
# Copyright (C) 2015 Nathaniel Smith <njs@pobox.com>
# See file LICENSE.txt for license information.

import numpy as np

def stacklast(*arrs):
    arrs = [np.asarray(arr)[..., np.newaxis] for arr in arrs]
    return np.concatenate(arrs, axis=-1)

def test_stacklast():
    assert np.array_equal(stacklast([1, 2, 3], [4, 5, 6]),
                          [[1, 4], [2, 5], [3, 6]])
    assert stacklast(np.zeros((3, 4)), np.zeros((3, 4))).shape == (3, 4, 2)

# Using color conventions: degrees 0-360
def color_cart2polar(a, b):
    h_rad = np.arctan2(b, a)
    h = np.rad2deg(h_rad) % 360
    r = np.hypot(a, b)
    return (r, h)

def test_color_cart2polar():
    assert np.allclose(color_cart2polar(0, 1), (1, 90))
    assert np.allclose(color_cart2polar(0, 10), (10, 90))
    assert np.allclose(color_cart2polar(-1, 0), (1, 180))

    assert np.allclose(color_cart2polar([0, 0, -1], [1, 10, 0]),
                       ([1, 10, 1], [90, 90, 180]))

def color_polar2cart(r, h):
    h_rad = np.deg2rad(h)
    return (r * np.cos(h_rad), r * np.sin(h_rad))

def test_color_polar2cart():
    assert np.allclose(color_polar2cart(1, 90), (0, 1))
    assert np.allclose(color_polar2cart(10, 90), (0, 10))
    assert np.allclose(color_polar2cart(1, 180), (-1, 0))

    assert np.allclose(color_polar2cart([1, 10, 1], [90, 90, 180]),
                       ([0, 0, -1], [1, 10, 0]))



# This file is part of colorspacious
# Copyright (C) 2014-2015 Nathaniel Smith <njs@pobox.com>
# See file LICENSE.txt for license information.

import numpy as np

# test_vector is like
#   [(gold_a, gold_b),
#    (gold_a, gold_b),
#    ...
#   ]

def check_conversion(forward, backward, test_vector,
                     a_min=0, a_max=100,
                     b_min=0, b_max=100,
                     # We use different precisions for tests against gold
                     # standard test vectors collected elsewhere, versus
                     # checks of our internal consistency. This is because
                     # other people's codebases introduce all kinds of
                     # rounding error -- they precompute inverse matrices and
                     # then round them before using, they replace fractions
                     # with decimals rounded to like 4 significant digits,
                     # etc. Internally, though, we strive for higher
                     # standards.
                     gold_rtol=1e-3, internal_rtol=1e-8,
                     **kwargs):
    a_min = np.asarray(a_min)
    a_max = np.asarray(a_max)
    b_min = np.asarray(b_min)
    b_max = np.asarray(b_max)

    def check_one(one_a, one_b):
        conv_b = forward(one_a, **kwargs)
        assert np.allclose(conv_b, one_b, rtol=gold_rtol)
        conv_a = backward(one_b, **kwargs)
        assert np.allclose(conv_a, one_a, rtol=gold_rtol)

    for a, b in test_vector:
        check_one(a, b)

    all_a = np.asarray([pair[0] for pair in test_vector])
    all_b = np.asarray([pair[1] for pair in test_vector])
    check_one(all_a, all_b)

    # make sure broadcasting / high-dimensionality conversion works
    all_a_3d = all_a[np.newaxis, ...]
    all_b_3d = all_b[np.newaxis, ...]
    check_one(all_a_3d, all_b_3d)

    # We use a stricter tolerance here, because
    r = np.random.RandomState(0)
    rand_a = r.uniform(size=(2, 2, 2, 2, 3))
    rand_a *= (a_max - a_min)
    rand_a += a_min
    assert np.allclose(backward(forward(rand_a, **kwargs), **kwargs),
                       rand_a,
                       rtol=internal_rtol)
    rand_b = r.uniform(size=(2, 2, 2, 2, 3))
    rand_b *= (b_max - b_min)
    rand_b += b_min
    assert np.allclose(forward(backward(rand_b, **kwargs), **kwargs),
                       rand_b,
                       rtol=internal_rtol)



# This file is part of colorspacious
# Copyright (C) 2014-2015 Nathaniel Smith <njs@pobox.com>
# See file LICENSE.txt for license information.

# Basic colorspaces: conversions between sRGB, XYZ, xyY, CIELab

import numpy as np

from .util import stacklast, color_cart2polar, color_polar2cart
from .illuminants import as_XYZ100_w
from .testing import check_conversion

################################################################
# sRGB <-> sRGB-linear <-> XYZ100
################################################################

# https://en.wikipedia.org/wiki/SRGB#The_reverse_transformation
def C_linear(C_srgb):
    out = np.empty(C_srgb.shape, dtype=float)
    linear_portion = (C_srgb < 0.04045)
    a = 0.055
    out[linear_portion] = C_srgb[linear_portion] / 12.92
    out[~linear_portion] = ((C_srgb[~linear_portion] + a) / (a + 1)) ** 2.4
    return out

def C_srgb(C_linear):
    out = np.empty(C_linear.shape, dtype=float)
    linear_portion = (C_linear <= 0.0031308)
    a = 0.055
    out[linear_portion] = C_linear[linear_portion] * 12.92
    out[~linear_portion] = (1+a) * C_linear[~linear_portion] ** (1/2.4) - a
    return out

XYZ100_to_sRGB1_matrix = np.array([
    # This is the exact matrix specified in IEC 61966-2-1:1999
    [ 3.2406, -1.5372, -0.4986],
    [-0.9689,  1.8758,  0.0415],
    [ 0.0557, -0.2040,  1.0570],
    ])

# Condition number is 4.3, inversion is safe:
sRGB1_to_XYZ100_matrix = np.linalg.inv(XYZ100_to_sRGB1_matrix)

def XYZ100_to_sRGB1_linear(XYZ100):
    """Convert XYZ to linear sRGB, where XYZ is normalized so that reference
    white D65 is X=95.05, Y=100, Z=108.90 and sRGB is on the 0-1 scale. Linear
    sRGB has a linear relationship to actual light, so it is an appropriate
    space for simulating light (e.g. for alpha blending).

    """
    XYZ100 = np.asarray(XYZ100, dtype=float)
    # this is broadcasting matrix * array-of-vectors, where the vector is the
    # last dim
    RGB_linear = np.einsum("...ij,...j->...i", XYZ100_to_sRGB1_matrix, XYZ100 / 100)
    return RGB_linear

def sRGB1_linear_to_sRGB1(sRGB1_linear):
    return C_srgb(np.asarray(sRGB1_linear, dtype=float))

def sRGB1_to_sRGB1_linear(sRGB1):
    """Convert sRGB (as floats in the 0-to-1 range) to linear sRGB."""
    sRGB1 = np.asarray(sRGB1, dtype=float)
    sRGB1_linear = C_linear(sRGB1)
    return sRGB1_linear

def sRGB1_linear_to_XYZ100(sRGB1_linear):
    sRGB1_linear = np.asarray(sRGB1_linear, dtype=float)
    # this is broadcasting matrix * array-of-vectors, where the vector is the
    # last dim
    XYZ100 = np.einsum("...ij,...j->...i", sRGB1_to_XYZ100_matrix, sRGB1_linear)
    XYZ100 *= 100
    return XYZ100

def test_sRGB1_to_sRGB1_linear():
    from .gold_values import sRGB1_sRGB1_linear_gold
    check_conversion(sRGB1_to_sRGB1_linear, sRGB1_linear_to_sRGB1,
                     sRGB1_sRGB1_linear_gold,
                     a_max=1, b_max=1)

def test_sRGB1_linear_to_XYZ100():
    from .gold_values import sRGB1_linear_XYZ100_gold
    check_conversion(sRGB1_linear_to_XYZ100, XYZ100_to_sRGB1_linear,
                     sRGB1_linear_XYZ100_gold,
                     a_max=1, b_max=100)

################################################################
# XYZ <-> xyY
################################################################

# These functions work identically for both the 0-100 and 0-1 versions of
# XYZ/xyY.
def XYZ_to_xyY(XYZ):
    XYZ = np.asarray(XYZ, dtype=float)
    norm = np.sum(XYZ, axis=-1, keepdims=True)
    xy = XYZ[..., :2] / norm
    return np.concatenate((xy, XYZ[..., 1:2]), axis=-1)

def xyY_to_XYZ(xyY):
    xyY = np.asarray(xyY, dtype=float)
    x = xyY[..., 0]
    y = xyY[..., 1]
    Y = xyY[..., 2]
    X = Y / y * x
    Z = Y / y * (1 - x - y)
    return stacklast(X, Y, Z)

_XYZ100_to_xyY100_test_vectors = [
    ([10, 20, 30], [ 10. / 60,  20. / 60, 20]),
    ([99, 98,  3], [99. / 200, 98. / 200, 98]),
    ]

_XYZ1_to_xyY1_test_vectors = [
    ([0.10, 0.20, 0.30], [ 0.10 / 0.60,  0.20 / 0.60, 0.20]),
    ([0.99, 0.98, 0.03], [0.99 / 2.00, 0.98 / 2.00, 0.98]),
    ]

def test_XYZ_to_xyY():
    check_conversion(XYZ_to_xyY, xyY_to_XYZ,
                     _XYZ100_to_xyY100_test_vectors, b_max=[1, 1, 100])

    check_conversion(XYZ_to_xyY, xyY_to_XYZ,
                     _XYZ1_to_xyY1_test_vectors, b_max=[1, 1, 1])

################################################################
# XYZ100 <-> CIEL*a*b*
################################################################

# https://en.wikipedia.org/wiki/Lab_color_space#CIELAB-CIEXYZ_conversions
def _f(t):
    out = np.empty(t.shape, dtype=float)
    linear_portion = (t < (6. / 29) ** 3)
    out[linear_portion] = ((1. / 3) * (29. / 6) ** 2 * t[linear_portion]
                           + 4. / 29)
    out[~linear_portion] = t[~linear_portion] ** (1. / 3)
    return out

def XYZ100_to_CIELab(XYZ100, XYZ100_w):
    XYZ100 = np.asarray(XYZ100, dtype=float)
    XYZ100_w = as_XYZ100_w(XYZ100_w)

    fXYZ100_norm = _f(XYZ100 / XYZ100_w)
    L = 116 * fXYZ100_norm[..., 1:2] - 16
    a = 500 * (fXYZ100_norm[..., 0:1] - fXYZ100_norm[..., 1:2])
    b = 200 * (fXYZ100_norm[..., 1:2] - fXYZ100_norm[..., 2:3])
    return np.concatenate((L, a, b), axis=-1)

def _finv(t):
    linear_portion = (t <= 6. / 29)
    out = np.select([linear_portion, ~linear_portion],
                    [3 * (6. / 29) ** 2 * (t - 4. / 29),
                     t ** 3])
    return out

def CIELab_to_XYZ100(CIELab, XYZ100_w):
    CIELab = np.asarray(CIELab, dtype=float)
    XYZ100_w = as_XYZ100_w(XYZ100_w)

    L = CIELab[..., 0]
    a = CIELab[..., 1]
    b = CIELab[..., 2]
    X_w = XYZ100_w[..., 0]
    Y_w = XYZ100_w[..., 1]
    Z_w = XYZ100_w[..., 2]

    l_piece = 1. / 116 * (L + 16)
    X = X_w * _finv(l_piece + 1. / 500 * a)
    Y = Y_w * _finv(l_piece)
    Z = Z_w * _finv(l_piece - 1. / 200 * b)

    return stacklast(X, Y, Z)

def test_XYZ100_to_CIELab():
    from .gold_values import XYZ100_CIELab_gold_D65, XYZ100_CIELab_gold_D50

    check_conversion(XYZ100_to_CIELab, CIELab_to_XYZ100,
                     XYZ100_CIELab_gold_D65,
                     # Stick to randomized values in the mid-range to avoid
                     # hitting negative luminances
                     b_min=[10, -30, -30], b_max=[90, 30, 30],
                     XYZ100_w="D65")

    check_conversion(XYZ100_to_CIELab, CIELab_to_XYZ100,
                     XYZ100_CIELab_gold_D50,
                     # Stick to randomized values in the mid-range to avoid
                     # hitting negative luminances
                     b_min=[10, -30, -30], b_max=[90, 30, 30],
                     XYZ100_w="D50")

    XYZ100_1 = np.asarray(XYZ100_CIELab_gold_D65[0][0])
    CIELab_1 = np.asarray(XYZ100_CIELab_gold_D65[0][1])

    XYZ100_2 = np.asarray(XYZ100_CIELab_gold_D50[1][0])
    CIELab_2 = np.asarray(XYZ100_CIELab_gold_D50[1][1])

    XYZ100_mixed = np.concatenate((XYZ100_1[np.newaxis, :],
                                   XYZ100_2[np.newaxis, :]))
    CIELab_mixed = np.concatenate((CIELab_1[np.newaxis, :],
                                   CIELab_2[np.newaxis, :]))

    XYZ100_w_mixed = np.row_stack((as_XYZ100_w("D65"), as_XYZ100_w("D50")))

    assert np.allclose(XYZ100_to_CIELab(XYZ100_mixed, XYZ100_w=XYZ100_w_mixed),
                       CIELab_mixed, rtol=0.001)
    assert np.allclose(CIELab_to_XYZ100(CIELab_mixed, XYZ100_w=XYZ100_w_mixed),
                       XYZ100_mixed, rtol=0.001)

################################################################
# CIELab <-> CIELCh
################################################################

def CIELab_to_CIELCh(CIELab):
    CIELab = np.asarray(CIELab)
    L = CIELab[..., 0]
    a = CIELab[..., 1]
    b = CIELab[..., 2]
    C, h = color_cart2polar(a, b)
    return stacklast(L, C, h)

def CIELCh_to_CIELab(CIELCh):
    CIELCh = np.asarray(CIELCh)
    L = CIELCh[..., 0]
    C = CIELCh[..., 1]
    h = CIELCh[..., 2]
    a, b = color_polar2cart(C, h)
    return stacklast(L, a, b)
